# This migration tool converts OSQA data to Askbot, optimized for big databases
# Copyright (C) 2016  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function

from . import django_integration
from . import utils


class Migrator:

    def __init__(self, osqa_conn, askbot_path, askbot_env):
        # Connect to the OSQA database
        self.osqa = utils.Connection(osqa_conn)

        # Load Askbot
        self.django = django_integration.load_django(askbot_path, askbot_env)
        self.models = django_integration.load_models()
        self.openid_models = django_integration.load_openid_models()

    def set_next_id(self, table, field, value):
        """Set the next ID for a sequence"""
        self.askbot_query(
            "SELECT setval("
                "(SELECT pg_get_serial_sequence(%s, %s)), %s"
            ")"
        ";", table, field, value)

    def askbot_query(self, query, *params):
        """Execute a query against Askbot"""
        with self.django.db.connection.cursor() as cursor:
            cursor.execute(query, params)
            return cursor.fetchall()

    def migrate(self):
        """Start the migration process"""
        self.migrate_users()
        self.migrate_user_logins()
        self.migrate_tags()
        self.migrate_nodes("question")
        self.migrate_nodes("answer")
        self.migrate_nodes("comment")

    def migrate_users(self):
        """Migrate users to Askbot"""
        migrate_fields_user = [
            "id", "email", "username", "first_name", "last_name", "last_login",
            "date_joined",
        ]
        migrate_fields_profile = [
            "website", "email_isvalid", "real_name", "date_of_birth",
            "location", "last_seen",
        ]

        # Load existing users
        existing = [u[0] for u in self.askbot_query(
            "SELECT id FROM auth_user;"
        )]

        last_id = 0
        tbl = "auth_user a, forum_user f WHERE a.id = f.user_ptr_id"
        for user in self.osqa.table_items(tbl, what="users", order="id"):
            last_id = user["id"]

            # Check if the user already exists
            if user["id"] in existing:
                continue

            # Create a new User object
            new = self.models.User()
            new.avatar_type = "g"  # Gravatar

            # Copy some fields
            for field in migrate_fields_user:
                setattr(new, field, user[field])

            # Update some other details of the user
            new.set_unusable_password()
            if user["is_superuser"]:
                new.set_admin_status()

            # Save the user
            new.save()

            # Update the user profile
            for field in migrate_fields_profile:
                setattr(new, field, user[field])

            # Update the about text
            new.update_localized_profile(about=user["about"])

            # Save the user again
            new.save()

        # Update the next ID value
        self.set_next_id("auth_user", "id", last_id + 1)

    def migrate_user_logins(self):
        """Migrate the login table"""
        # Load existing users
        existing = [u[0] for u in self.askbot_query(
            "SELECT id FROM django_authopenid_userassociation;"
        )]

        last_id = 0
        t = "forum_authkeyuserassociation"
        for login in self.osqa.table_items(t, what="user logins", order="id"):
            last_id = login["id"]

            # Check if the login entry already exists
            if login["id"] in existing:
                continue

            # Migrate login data
            assoc = self.openid_models.UserAssociation()
            assoc.id = login["id"]
            assoc.openid_url = login["key"]
            assoc.user_id = login["user_id"]
            assoc.provider_name = login["provider"]
            assoc.last_used_timestamp = login["added_at"]
            try:
                assoc.save()
            except self.django.db.utils.IntegrityError:
                # OSQA forgot to delete a record -- don't add it
                pass

        # Update the next ID value
        self.set_next_id("django_authopenid_userassociation", "id", last_id + 1)

    def migrate_tags(self):
        """Migrate tags"""
        existing = [n[0] for n in self.askbot_query("SELECT id FROM tag;")]

        last_id = 0
        for tag in self.osqa.table_items("forum_tag", what="tags", order="id"):
            last_id = tag["id"]

            if tag["id"] in existing:
                continue

            new = self.models.Tag()
            new.id = tag["id"]
            new.name = tag["name"]
            new.created_by_id = tag["created_by_id"]
            new.used_count = tag["used_count"]
            new.save()

        self.set_next_id("tag", "id", last_id + 1)

    def migrate_nodes(self, node):
        """Migrate questions"""
        # Load existing nodes
        existing = [n[0] for n in self.askbot_query(
            "SELECT id FROM askbot_post WHERE post_type = %s;", node
        )]
        existing_threads = [n[0] for n in self.askbot_query(
            "SELECT id FROM askbot_thread;", node
        )]

        last_id = 0
        last_thread_id = 0
        table = "forum_node WHERE node_type = '%s'" % node
        for node in self.osqa.table_items(table, what="%ss" % node, order="id"):
            last_id = node["id"]

            # Check if the node already exists
            if node["id"] in existing:
                continue

            author = self.models.User.objects.get(id=node["author_id"])

            # If it's a question, create a new thread
            if node["node_type"] == "question":
                last_thread_id = node["id"]

                # Do this only if the thread doesn't exist
                if node["id"] not in existing_threads:
                    thread = self.models.Thread()
                    thread.id = node["id"]
                    thread.title = node["title"]

                    last_activity_by = node["author_id"]
                    if node["last_activity_by_id"]:
                        last_activity_by = node["last_activity_by_id"]

                    thread.last_activity_by_id = last_activity_by
                    thread.last_activity_at = node["last_activity_at"]

                    thread.view_count = node["extra_count"]

                    thread.tagnames = node["tagnames"]
                    thread.save()

            # Migrate post data
            post = self.models.Post()
            post.post_type = node["node_type"]
            post.id = node["id"]
            post.author_id = node["author_id"]
            post.text = node["body"]
            post.added_at = node["added_at"]

            # Attach parents
            if node["parent_id"]:
                post.parent_id = node["parent_id"]
            if node["node_type"] == "question":
                post.thread_id = node["id"]
            elif node["node_type"] == "answer":
                post.thread_id = node["parent_id"]
            elif node["node_type"] == "comment":
                post.thread_id = self.models.Post.objects \
                                     .get(id=node["parent_id"]) \
                                     .thread_id

            post.save()
            post.parse_and_save(author)

            # Mark answers as accepted if needed
            if node["marked"] and node["node_type"] == "answer":
                post.thread.accepted_answer = post
                post.endorsed = True
                post.endorsed_at = self.django.utils.timezone.now()
                post.thread.save()

            post.add_revision(
                author = author,
                revised_at = node["added_at"],
                text = node["body"],
                comment = "",
            )
